@@ -90,13 +90,13 @@ using namespace TagLib;
 
 /* Declaring the functions which should be accessible on the C side. */
 extern "C" {
-#define CAML_INTERNALS
 #include <caml/alloc.h>
 #include <caml/callback.h>
 #include <caml/fail.h>
 #include <caml/memory.h>
 #include <caml/misc.h>
 #include <caml/signals.h>
+
 #ifdef WIN32
 #include <caml/osdeps.h>
 #endif
@@ -170,17 +170,47 @@ CAMLprim value caml_taglib_init(value unit) {
   CAMLreturn(Val_unit);
 }
 
-#define Taglib_file_val(v) ((File *)v)
+#define Taglib_file_val(v) (*((File **)Data_abstract_val(v)))
+
+static inline value value_of_taglib_file(value v, File *f) {
+  v = caml_alloc(1, Abstract_tag);
+  *((File **)Data_abstract_val(v)) = f;
+  return v;
+}
+
 #define Taglib_tag_val(v) (*((Tag **)Data_custom_val(v)))
-#define Taglib_audioproperties_val(v) ((AudioProperties *)v)
+
+// tag are implemented as custom val in the id3v2 case so we need
+// to be consistent here.
+static struct custom_operations tag_ops = {
+    (char *)"ocaml_taglib_tag", custom_finalize_default,
+    custom_compare_default,     custom_hash_default,
+    custom_serialize_default,   custom_deserialize_default};
+
+static inline value value_of_taglib_tag(value v, Tag *tag) {
+  v = caml_alloc_custom(&tag_ops, sizeof(Tag *), 1, 0);
+  Taglib_tag_val(v) = tag;
+  return v;
+}
+
+#define Taglib_audioproperties_val(v)                                          \
+  (*((AudioProperties **)Data_abstract_val(v)))
+
+static inline value value_of_taglib_audioproperties(value v,
+                                                    AudioProperties *f) {
+  v = caml_alloc(1, Abstract_tag);
+  *((AudioProperties **)Data_abstract_val(v)) = f;
+  return v;
+}
 
 CAMLprim value caml_taglib_file_new(value type, value name) {
   CAMLparam2(name, type);
+  CAMLlocal1(ret);
 
   File *f = NULL;
 
 #ifdef WIN32
-  const wchar_t *filename = caml_stat_strdup_to_os(String_val(name));
+  wchar_t *filename = caml_stat_strdup_to_os(String_val(name));
 #else
   char *filename = strdup(String_val(name));
 #endif
@@ -238,7 +268,7 @@ CAMLprim value caml_taglib_file_new(value type, value name) {
     caml_raise_constant(*caml_named_value("taglib_exn_invalid_file"));
   }
 
-  CAMLreturn((value)f);
+  CAMLreturn(value_of_taglib_file(ret, f));
 }
 
 CAMLprim value caml_taglib_file_free(value f) {
@@ -249,14 +279,6 @@ CAMLprim value caml_taglib_file_free(value f) {
   CAMLreturn(Val_unit);
 }
 
-/* Do nothing file file tags. */
-static void finalize_file_tag(value t) { return; }
-
-static struct custom_operations file_tag_ops = {
-    (char *)"ocaml_taglib_file_tag", finalize_file_tag,
-    custom_compare_default,          custom_hash_default,
-    custom_serialize_default,        custom_deserialize_default};
-
 CAMLprim value caml_taglib_file_tag(value f) {
   CAMLparam1(f);
   CAMLlocal1(ret);
@@ -265,20 +287,18 @@ CAMLprim value caml_taglib_file_tag(value f) {
   if (t == NULL)
     caml_raise_constant(*caml_named_value("taglib_exn_not_found"));
 
-  ret = caml_alloc_custom(&file_tag_ops, sizeof(Tag *), 1, 0);
-  Taglib_tag_val(ret) = t;
-
-  CAMLreturn(ret);
+  CAMLreturn(value_of_taglib_tag(ret, t));
 }
 
 CAMLprim value caml_taglib_file_audioproperties(value f) {
   CAMLparam1(f);
+  CAMLlocal1(ret);
   AudioProperties *p = Taglib_file_val(f)->audioProperties();
 
   if (p == NULL)
     caml_raise_constant(*caml_named_value("taglib_exn_not_found"));
 
-  CAMLreturn((value)p);
+  CAMLreturn(value_of_taglib_audioproperties(ret, p));
 }
 
 CAMLprim value caml_taglib_file_save(value f) {
@@ -298,7 +318,7 @@ CAMLprim value caml_taglib_tag_get_string(value t, value name) {
   CAMLlocal1(ans);
   const Tag *tag = Taglib_tag_val(t);
   const char *s = String_val(name);
-  String tmp = String::null;
+  String tmp = "";
 
   if (!strcmp(s, "title"))
     tmp = tag->title();
@@ -313,7 +333,7 @@ CAMLprim value caml_taglib_tag_get_string(value t, value name) {
   else
     caml_failwith("Invalid value");
 
-  if (tmp == String::null)
+  if (tmp.isEmpty())
     caml_raise_constant(*caml_named_value("taglib_exn_not_found"));
 
   ans = caml_copy_string(tmp.toCString(bool(true)));
